///////////////////////////////////////////////////////////////////////////////
//                         I2C controller                                    //
//                       Jesús Arias (2022)                                  //
//                                                                           //
// Public Domain code (bugs included).                                       //
///////////////////////////////////////////////////////////////////////////////


module i2c_master (
	input clk,		// master clock
	input [10:0]d,	// Input data bus
	output [9:0]q,	// Output data
	input wr,		// Write data/control
	input wrdivi,	// Write clock divider
	
	output scl_out,	// Bus signals
	input  scl,
	output sda_out,
	input  sda
);

// State names
parameter IDLE		= 4'b0000;
parameter START1	= 4'b0001;
parameter START2	= 4'b0010;
parameter START3	= 4'b0011;
parameter DATA0		= 4'b0100; // ST[2] == 1
parameter DATA1		= 4'b0101; // ST[2] == 1
parameter DATA2		= 4'b0110; // ST[2] == 1
parameter DATA3		= 4'b0111; // ST[2] == 1
parameter STOP1		= 4'b1000; // 			ST[3] == 1
parameter STOP2		= 4'b1001; // 			ST[3] == 1
parameter STOP3		= 4'b1010; // 			ST[3] == 1

// Clock divider
reg [6:0]divider=7'h7F;
reg [6:0]divcnt=0;

wire pulse;
assign pulse=(divcnt==0);
always @(posedge clk) begin
	if (wrdivi) divider<=d[6:0];
	divcnt<=(wr & (st==IDLE)) ? 1 : (pulse ? divider : divcnt-1);
end

// Shifter & ancillary

reg [8:0]sh;	// Shifter
always @(posedge clk) 
	if (wr) sh<={d[7:0],d[8]};
	else if ((st==7) & pulse) sh<={sh[7:0],sdasamp};

//always @(posedge clk) 
assign sda_out = st[2]? sh[8] : ((st==IDLE)|(st==START1)|(st==STOP3));


reg start=0;
reg stop=0;
always @(posedge clk) if (wr) {stop,start}<=d[10:9];

reg sdasamp;	// input FF
always @(posedge clk) if ((st==DATA2)&pulse) sdasamp<=sda;

reg [3:0]bitcnt=4'hf;	// Bit counter
wire busy;
assign busy=(st[2]&(bitcnt!=0)) | (st==START1) | (st==START2) | (st==START3) | st[3];

always @(posedge clk) 
	if (wr) bitcnt<=st[2] ? 4'h9 : 4'h0;  
	else if((st==7)&pulse) bitcnt<=bitcnt-1;
	
assign q={busy,sh[0],sh[8:1]};
//always @(posedge clk) 
assign scl_out=(st==IDLE)|(st==START1)|(st==START2)|(st==DATA1)|(st==DATA2)|(st==STOP2)|(st==STOP3);

// State machine

reg [3:0]st=IDLE;	

always @(posedge clk)
begin
  if (pulse)
	case (st)
	IDLE:	if (start) st<=START1;
	START1: st<=START2;
	START2: st<=START3;
	START3: st<=DATA0;
	DATA0:	if (stop) st<=STOP1; else if (busy) st<=DATA1;
	DATA1:	if (scl) st<=DATA2;
	DATA2:	st<=DATA3;
	DATA3:	st<=DATA0;
	STOP1:	st<=STOP2;
	STOP2:	st<=STOP3;
	STOP3:	st<=IDLE;
	default: st<=IDLE;
	endcase
end



endmodule
